@@ -23,6 +23,7 @@
 #include <vector>
 
 // #include "flashinfer/trtllm/common/Dtype.h"
+#include "trtllmGen_bmm_export/Enums.h"
 #include "trtllmGen_bmm_export/trtllm/gen/DtypeDecl.h"
 
 namespace tensorrt_llm {
@@ -38,6 +39,8 @@ struct TrtllmGenBatchedGemmRunnerOptions {
   bool transposeMmaOutput{false};
   int32_t tileSize{8};
   int32_t epilogueTileM{128};
+  bool useShuffledMatrixA{false};
+  batchedGemm::gemm::MatrixLayout weightLayout{batchedGemm::gemm::MatrixLayout::MajorK};
 };
 
 class TrtllmGenBatchedGemmRunner {
@@ -48,8 +51,9 @@ class TrtllmGenBatchedGemmRunner {
                                                std::vector<int32_t> const& batchedTokens,
                                                int32_t numTokens, int32_t numBatches,
                                                int32_t maxNumCtasInBatchDim,
-                                               std::optional<int32_t> configIndex = std::nullopt);
+                                               int32_t configIndex) const;
 
+  // Generic GEMM interface
   void run(int32_t m, int32_t n, int32_t k, std::vector<int32_t> const& batchedTokens,
            int32_t numTokens, int32_t numBatches, int32_t maxNumCtasInBatchDim, void const* a,
            void const* sfA, void const* b, void const* sfB, void const* perTokensSfA,
@@ -57,20 +61,20 @@ class TrtllmGenBatchedGemmRunner {
            void* outSfC, int32_t const* routeMap, int32_t const* totalNumPaddedTokens,
            int32_t const* ctaIdxXyToBatchIdx, int32_t const* ctaIdxXyToMnLimit,
            int32_t const* numNonExitingCtas, void* workspace, CUstream stream, int device,
-           std::optional<int32_t> configIndex = std::nullopt);
+           int32_t configIndex);
 
+  // NVFP4 per-block scaling GEMM
   void run(int32_t m, int32_t n, int32_t k, std::vector<int32_t> const& batchedTokens,
            void const* a, void const* sfA, void const* b, void const* sfB, void* c, void* outSfC,
-           void* workspace, CUstream stream, int device,
-           std::optional<int32_t> configIndex = std::nullopt);
+           void* workspace, CUstream stream, int device, int32_t configIndex);
 
+  // FP8 per-tensor scaling GEMM
   void run(int32_t m, int32_t n, int32_t k, std::vector<int32_t> const& batchedTokens,
            void const* a, void const* b, float const* scaleC, float const* scaleGateC, void* c,
-           void* workspace, CUstream stream, int device,
-           std::optional<int32_t> configIndex = std::nullopt);
+           void* workspace, CUstream stream, int device, int32_t configIndex);
 
   // Get the list of configs that passed the validation based on the constructor options
-  [[nodiscard]] std::vector<int32_t> getPassingConfigIndices() const {
+  [[nodiscard]] std::vector<int64_t> getPassingConfigIndices() const {
     return mPassingConfigIndices;
   }
 
@@ -86,14 +90,18 @@ class TrtllmGenBatchedGemmRunner {
                                                    int32_t numTokens, int32_t numBatches,
                                                    int32_t maxNumCtasInBatchDim) const;
 
+  [[nodiscard]] bool isValidConfigIndex(int32_t configIndex, int32_t m, int32_t n, int32_t k,
+                                        std::vector<int32_t> const& batchedTokens,
+                                        int32_t numTokens, int32_t numBatches,
+                                        int32_t maxNumCtasInBatchDim) const;
+
  private:
   void selectGemmConfig(int32_t m, int32_t n, int32_t k, std::vector<int32_t> const& batchedTokens,
                         int32_t numTokens, int32_t numBatches, int32_t maxNumCtasInBatchDim);
 
  private:
   TrtllmGenBatchedGemmRunnerOptions mOptions;
-  std::vector<int32_t> mPassingConfigIndices;
-  std::optional<int32_t> mSelectedConfigIndex;
+  std::vector<int64_t> mPassingConfigIndices;
 };
 }  // namespace kernels
 }  // namespace tensorrt_llm
